unit rkView;

//  rkView  2009 by Roy Magne Klever. All rights reserved
//
//  This file is not distributable without permission by Roy Magne Klever
//  WEB: www.rmklever.com
//  Mail: roymagne@rmklever.com
//
//  version 1.0

interface

uses
  Windows, Messages, WinTypes, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ComCtrls, StrUtils, ExtCtrls, Math, Stdctrls, Buttons,
  rkIntegerList;

const
  CM_REFRESH = WM_USER + 2000; // Custom Message...

type
  PRGB24 = ^TRGB24;
  TRGB24 = packed record
    B: Byte;
    G: Byte;
    R: Byte;
  end;
  TLine24 = array[0..0] of TRGB24;
  PLine24 = ^TLine24;

  TWMMouseWheel = record
    Msg: Cardinal;
    Keys: SmallInt;
    WheelDelta: SmallInt;
    case Integer of
      0: (
        XPos: Smallint;
        YPos: Smallint);
      1: (
        Pos: TSmallPoint;
        Result: Longint);
  end;

  TsvItemState = (svNormal, svSelected, svHot, svHotSelected);

  // View events...
  TOnSelectingEvent = procedure(Sender: TObject; Count: Integer) of
    object;
  TOnHeaderPaintEvent = procedure(Sender: TObject; Canvas: TCanvas; Header: TRect;
    Offset, Active: Integer; State: TsvItemState; Columns: Array of Integer) of object;
  TOnCellPaintEvent = procedure(Sender: TObject; Canvas: TCanvas; Cell: TRect;
    IdxA, Idx: Integer; State: TsvItemState) of object;
  TOnListPaintEvent = procedure(Sender: TObject; Canvas: TCanvas; Cell: TRect;
    IdxA, Idx: Integer; State: TsvItemState; Columns: array of Integer) of object;
  TOnCellSelectEvent = procedure(Sender: TObject; Canvas: TCanvas; Cell: TRect;
    Idx: Integer; var Selected: Boolean) of object;
  TOnCellHitEvent = procedure(Sender: TObject; Canvas: TCanvas;
    Idx, x, y: Integer; var Selected: Boolean) of object;
  TOnHeaderClickEvent = procedure(Sender: TObject; Column: Integer) of object;
  TOnHintShowEvent = procedure(Sender: Tobject; idx, column: Integer;
    var text: String; var Show: Boolean) of object;

  /// Main class
  TrkView = class(TCustomControl)
  private
    FFocused: Boolean;
    FModified: Boolean;
    FMouseItem: Integer;
    FMouseAtEnd: Boolean;
    FDownTick: DWord;
    FDownX: Integer;
    FDownY: Integer;

    // Header management
    FHeaderHot: Integer;
    FHeaderState: TsvItemState;
    FDragColumn: Integer;
    FDragging: Boolean;
    FDragOffs: Integer;
    FDragStartX: Integer;
    FDragStartSize: Integer;
    InHeader: Boolean;
    FHeaderVisible: Boolean;

    // Cell data
    FCellWidth: Integer;
    FCellHeight: Integer;
    FCellOffset: Integer;
    FCellSpace: Integer;
    FCellSpaceX: Integer;
    FCellXOffset: Integer;
    FCtrlSize: Integer;   // Size of VertScrollVar
    FHeaderHeight: Integer;
    FCellAuto: Boolean;
    CellColSize: Integer;

    // Selection
    SelScroll: boolean;
    SelStart, SelEnd: tPoint;
    viewDeltaScroll, ScrollOffset: Integer;
    viewDeltaScrollX, ScrollOffsetX: Integer;
    maxXScroll: Integer;
    SelTimer: TTimer;
    Ra, Ga, Ba: Array[0..255] of Byte;

    // Painting
    FMainBmp: TBitmap;
    FFont: TFont;
    FBackColor: TColor;
    FLinesColor: TColor;
    FColorSelection: TRGB24;

    // Scrollmode
    SCUp: Boolean;
    SCY, SCPos, SCValue: Integer;

    // Scrollbar
    sbVert: TScrollBar;
    sbHorz: TScrollBar;
    FScrollValOld: integer;
    FScrollValNew: integer;
    FScrollPos: Integer;
    Range: Integer; // Needed for prop calc
    hsbVisible: Boolean;
    vsbVisible: Boolean;
    YAdjOff: Integer;

    // Parts of component  Easier to maintain scrollbars :)
    pnlVert, pnlHorz, pnlDummy: TPanel;

    dblClicked: boolean;

    FSelected: Integer;
    FUseAsList : Boolean;
    FCellSelect: Boolean;
    FCenterView: Boolean;
    FHotTracking: Boolean;
    FPaintGrid: Boolean;
    FMulti: Boolean;         // Multi selection

    FListWidth: Integer;
    FColumns: Array of Integer;
    FHintColumn: Integer;

    FLastIndex: Integer;
    FNrSelected: Integer;

    FPopupID: Integer;

    FBorderStyle: TBorderStyle;
    FFullRepaint: Boolean;   // ?
    FLocked: Boolean;        // ?

    FOnResize: TNotifyEvent;
    FOnEnter: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnMouseDown: TMouseEvent;
    FOnMouseUp: TMouseEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnDblClick: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FOnMouseWheel: TMouseWheelEvent;

    FAlignment: TAlignment;

    function GetColumns: String;
    procedure SetColumns(Value: String);
    procedure SetUseAsList(Value: Boolean);
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMRefresh(var Message: TMessage); message CM_REFRESH;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetFont(Value: TFont);
    procedure WMGetDlgCode(var message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMIsToolControl(var Message: TMessage); message CM_ISTOOLCONTROL;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message
      WM_WINDOWPOSCHANGED;
    procedure WMMouseWheel(var Msg: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure OnVertScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos:
      Integer);
    procedure OnHorzScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos:
      Integer);
    procedure CalcSelectionSimple;
    procedure CalcSelection;
  protected
    FOnSelecting: TOnSelectingEvent;
    FOnCellSelect: TOnCellSelectEvent;
    FOnCellHit: TOnCellHitEvent;
    FOnHeaderPaint: TOnHeaderPaintEvent;
    FOnHeaderClick: TOnHeaderClickEvent;
    FOnCellPaint: TOnCellPaintEvent;
    FOnListPaint: TOnListPaintEvent;
    FOnHintShow: TOnHintShowEvent;
    FCanFocused: boolean;

    procedure PaintSimple;
    procedure PaintWindow(DC: HDC); override;
    procedure WMERASEBKGND(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMSETFOCUS(var Message: TWMSETFOCUS); message WM_SETFOCUS;
    procedure WMKILLFOCUS(var Message: TWMKILLFOCUS); message WM_KILLFOCUS;

    procedure SetCanFocused(Value: Boolean);
    procedure SetSelColor(Value: TColor);
    function GetSelColor: TColor;

    function IsFocused: Boolean;
    procedure SimpleMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:
      Integer);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:
      Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure DblClick; override;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMKeyUp(var Message: TWMKeyUp); message WM_KEYUP;
    function SimpleDoKey(var Key: Word; Shift: TShiftState): Boolean;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure OnSelScroll(Sender: Tobject);
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
  public
    IsEditing: Boolean;
    BugStr: String;
    BugCount: Integer;
    Items: TIntList;
    Selection: TIntList;
    IdxItem: Integer;
    LastIdxItem: Integer;
    HotIdx: Integer;
    ViewIdx: Integer;
    // Selection
    Selecting: Boolean;
    FSelRect, FPrevSel, SelRect: TRect;
    //Selected: Integer;
    SelX, SelY: Integer;
    HotX, HotY: Integer;
    // View data
    ImgWidth, ImgHeight, ViewColumns, ViewRows: Integer;
    ScrollMode: Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HzScrollbar: Boolean;
    procedure CalcViewSimple;
    procedure CalcHorzSB;
    procedure CalcView(Clear: Boolean);
    procedure UpdateView;
    procedure Clear;
    function TextWidth(text: String): Integer;
    function GetColumnSize(Column: Integer): Integer;
    procedure SetInView(grp, idx: Integer);
    procedure SetAtTop(grp, idx: Integer);
    function GetItemRect(grp, idx: Integer): TRect;
    function GetItemAtXY(pt: TPoint): Integer;
    function IdxAtXY(pt: TPoint; HitTest: Boolean): integer;
    function ItemAtXY(pt: TPoint; HitTest: Boolean): integer;
    function ItemCloseToXY(X, Y: Integer): integer;
    function Scroll(delta, value: integer): integer;
    procedure Resize; override;
    property Selected: Integer read FSelected;
    property NumberSel: Integer read FNrSelected default 0;
    property Modified: Boolean read FModified;
  published
    // Published
    property Align;
    property Anchors;
    property Enabled;
    property PopupMenu;
    property ShowHint;
    property Hint;
    property Visible;
    property TabOrder;
    property Ctl3D;
    property ParentCtl3D;
    property DragCursor;
    property DragMode;

    // Cell data
    property CellWidth: Integer read FCellWidth write FCellWidth default 150;
    property CellHeight: Integer read FCellHeight write FCellHeight default 150;
    property CellOffset: Integer read FCellOffset write FCellOffset default 10;
    property CellSpace: Integer read FCellSpace write FCellSpace default 10;
    property HeaderHeight: Integer read FHeaderHeight write FHeaderHeight default 22;
    property CellAutoAdj: Boolean read FCellAuto write FCellAuto default False;
    property MultipleSelection: Boolean read FMulti write FMulti default False;
    property UseAsList: Boolean read FUseAsList write SetUseAsList default False;
    property CenterView: Boolean read FCenterView write FCenterView default False;
    property CellSelect: Boolean read FCellSelect write FCellSelect default False;
    property HotTracking: Boolean read FHotTracking write FHotTracking default False;
    property PaintGrid: Boolean read FPaintGrid write FPaintGrid default False;
    property HeaderVisible: Boolean read FHeaderVisible write FHeaderVisible default False;
    property Columns: String read GetColumns write SetColumns;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle
      default bsSingle;
    property Font: TFont read FFont write SetFont;
    property Color: TColor read FBackColor write FBackColor default clWindow;
    property ColorSel: TColor read GetSelColor write SetSelColor default
      clHighlight;
    property ColorLines: TColor read FLinesColor write FLinesColor default clSilver;
    property ParentColor;
    property TabStop;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnMouseMove: TMouseMoveEvent read FonMouseMove write FOnMouseMove;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnMouseWheel: TMouseWheelEvent read FOnMouseWheel write FOnMouseWheel;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter: TNotifyEvent read FOnEnter write FOnEnter;
    property OnExit: TNotifyEvent read FOnExit write FOnExit;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnStartDrag;
    property OnHintShow: TOnHintShowEvent read FOnHintShow write FOnHintShow;
    property OnSelecting: TOnSelectingEvent read FOnSelecting write FOnSelecting;
    property OnCellHit: TOnCellHitEvent read FOnCellHit write FOnCellHit;
    property OnCellSelect: TOnCellSelectEvent read FOnCellSelect write FOnCellSelect;
    property OnHeaderPaint: TOnHeaderPaintEvent read FOnHeaderPaint write FOnHeaderPaint;
    property OnHeaderClick: TOnHeaderClickEvent read FOnHeaderClick write FOnHeaderClick;
    property OnCellPaint: TOnCellPaintEvent read FOnCellPaint write FOnCellPaint;
    property OnListPaint: TOnListPaintEvent read FOnListPaint write FOnListPaint;
  end;

procedure Register;

implementation

constructor TrkView.Create(AOwner: TComponent);
var
  w, h: integer;
begin
  inherited;
  bugStr:= '';
  bugCount:= 0;
  w:= GetSystemMetrics(SM_CXVSCROLL); // Width of a vertical scrollbar...
  h:= GetSystemMetrics(SM_CXHSCROLL); // Width of a horizontal scrollbar...
  ControlStyle := ControlStyle + [csOpaque, csAcceptsControls, csReplicatable,
    csNeedsBorderPaint];
  Width := 185; // Init width
  Height := 82; //  ""  height
  // Cell data
  FCellWidth := 150;
  FCellHeight := 150;
  FCellOffset := 10;
  FCellSpace := 10;
  FCellSpaceX := 10;
  FCtrlSize := 0;
  FHeaderHeight:= 22;
  FDragColumn:= -1;
  FDragging:= False;
  FHeaderState:= svNormal;
  FHeaderHot:= -1;
  FAlignment:= taCenter;
  FBorderStyle:= bsSingle;
  Color:= clWindow;
  ColorSel:= clHighlight;
  ColorLines:= clSilver;
  FFullRepaint := True;
  FSelected:= -1;
  HotIdx:= -1;
  FPopupID := -1;
  FCanFocused := True;
  FHeaderVisible:= False;
  Items := TIntList.Create;
  Selection := TIntList.Create;
  FFont := TFont.Create;
  FFont.Name := 'Tahoma';
  FFont.Color := clBlack;
  FFont.Size := 8;
  FMainBmp := TBitmap.Create;
  FMainBmp.PixelFormat := pf24Bit;
  SelTimer := TTimer.Create(self);
  SelTimer.Enabled := false;
  SelTimer.Interval := 15;
  SelTimer.OnTimer := OnSelScroll;
  pnlHorz:= TPanel.Create(Self);
  pnlHorz.Parent := Self;
  pnlHorz.Align := alBottom;
  pnlHorz.BevelInner := TBevelCut(0);
  pnlHorz.BevelOuter := TBevelCut(0);
  pnlHorz.Height := h;
  pnlHorz.ParentBackground := false;
  pnlHorz.Visible := false;
  pnlVert:= TPanel.Create(Self);
  pnlVert.Parent := Self;
  pnlVert.Align := alRight;
  pnlVert.BevelInner := TBevelCut(0);
  pnlVert.BevelOuter := TBevelCut(0);
  pnlVert.Width := w;
  pnlVert.ParentBackground := false;
  pnlVert.Visible := false;
  pnlDummy:= TPanel.Create(pnlHorz);
  pnlDummy.Parent := pnlHorz;
  pnlDummy.Align := alRight;
  pnlDummy.BevelInner := TBevelCut(0);
  pnlDummy.BevelOuter := TBevelCut(0);
  pnlDummy.Width := w;
  pnlDummy.ParentBackground := false;
  pnlDummy.Visible := false;
  sbVert := TScrollBar.Create(pnlVert);
  sbVert.Parent := pnlVert;
  sbVert.Kind := sbVertical;
  sbVert.LargeChange := 1;
  sbVert.SmallChange := 1;
  sbVert.Align := alClient;
  sbVert.OnScroll := OnVertScroll;
  sbVert.TabStop := false;
  sbVert.Visible := true;
  sbHorz := TScrollBar.Create(pnlHorz);
  sbHorz.Parent := pnlHorz;
  sbHorz.Kind := sbHorizontal;
  sbHorz.LargeChange := 1;
  sbHorz.SmallChange := 1;
  sbHorz.Align := alClient;
  sbHorz.OnScroll := OnHorzScroll;
  sbHorz.TabStop := false;
  sbHorz.Visible := true;
  hsbVisible:= false;
  vsbVisible:= false;
  Selecting := false;
  SelStart.X := 0;
  SelStart.Y := 0;
  SelEnd.X := 0;
  SelEnd.Y := 0;
  SelScroll := false;
  ViewIdx := -1;
  CenterView:= False;
  HotTracking:= True;
  CellSelect:= True;
  PaintGrid:= False;
  FCellAuto:= False;
  FMulti := False;
  IsEditing:= False;
  FModified := False;
  Invalidate;
  ShowHint:= True;
end;

destructor TrkView.Destroy;
begin
  Items.Free;
  Selection.Free;
  FFont.Free;
  FMainBmp.Free;
  inherited;
end;

procedure TrkView.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrkView.PaintWindow(DC: HDC);
begin
  Canvas.Lock;
  try
    Canvas.Handle := DC;
    try
      FMainBmp.Canvas.Font:= FFont;
      PaintSimple;
    finally
      Canvas.Handle := 0;
    end;
  finally
    Canvas.Unlock;
  end;
end;

procedure TrkView.ReSize;
begin
  pnlDummy.Visible:= pnlVert.Visible;
  CalcView(False);
  if Assigned(FOnReSize) then FOnReSize(Self);
end;

function TrkView.GetColumns: String;
var
  i: Integer;
begin
  Result:= '';
  if High(FColumns) > 0 then begin
    for i := Low(FColumns) to High(FColumns) - 1 do
      Result:= Result + IntToStr(FColumns[i]) + ', ';
    Result:= Result + IntToStr(FColumns[High(FColumns)]);
  end else
  Result:= '';
end;

procedure TrkView.SetColumns(Value: String);
var
  i: Integer;
  s: TStringList;
begin
  s:= TStringList.Create;
  s.Text:= StringReplace(Value, ',', #13, [rfReplaceAll]);
  SetLength(FColumns, s.Count);
  i:= 0;
  FListWidth:= 0;
  while i < s.Count do begin
    FColumns[i]:= StrToIntDef(Trim(s[i]), 0);
    FListWidth:= FListWidth + FColumns[i];
    inc(i);
  end;
  s.Free;
end;

procedure TrkView.SetUseAsList(Value: Boolean);
begin
  FUseAsList:= Value;
  if FUseAsList then FCellWidth:= FListWidth;
end;

procedure TrkView.CMEnter(var Message: TCMEnter);
begin
  inherited;
  if not (csDesigning in ComponentState) then SetFocus;
  if Assigned(OnEnter) then  OnEnter(Self);
end;

procedure TrkView.CMExit(var Message: TCMExit);
begin
  if Assigned(OnExit) then OnExit(Self);
  inherited;
end;

procedure TrkView.CMRefresh(var Message: TMessage);
begin
  if (csDestroying in ComponentState) then Exit;
  if not (csDesigning in ComponentState) then PaintSimple;
  inherited;
end;

procedure TrkView.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TrkView.SetFont(Value: TFont);
begin
  if (FFont <> Value) then FFont.Assign(Value);
  Invalidate;
end;

procedure TrkView.WMGetDlgCode(var message: TWMGetDlgCode);
begin
  inherited;
  // Answer Delphi that this component wants to handle its own arrow key press:
  message.result := DLGC_WANTARROWS;
end;

procedure TrkView.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then
    RecreateWnd;
  inherited;
end;

procedure TrkView.CMIsToolControl(var Message: TMessage);
begin
  if not FLocked then
    Message.Result := 1;
end;

procedure TrkView.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  if fFullRepaint then
    Invalidate;
  inherited;
  if not (csLoading in ComponentState) then
    Resize;
end;

procedure TrkView.WMMouseWheel(var Msg: TWMMouseWheel);
var
  newPos: integer;
begin
  if not vsbVisible then Exit;
  if Msg.WheelDelta <> 0 then begin
    if Msg.WheelDelta < 0 then begin
      newPos := sbVert.Position + sbVert.SmallChange;
      if newPos > Range then newPos := Range;
      sbVert.Position := newPos;
    end else sbVert.Position := sbVert.Position - sbVert.SmallChange;
    Msg.Result := 1;
  end;
  if IsEditing then SetFocus;
  invalidate;
end;

procedure TrkView.OnVertScroll(Sender: TObject; ScrollCode: TScrollCode; var
  ScrollPos: Integer);
begin
  FScrollValOld:= FScrollValNew;
  FScrollValNew:= ScrollPos;
  FScrollPos:= FScrollValOld - FScrollValNew;
  if FScrollValOld <> FScrollValNew then begin
    if IsEditing then SetFocus;
    IsEditing:= False;
    Invalidate;
  end;
end;

procedure TrkView.OnHorzScroll(Sender: TObject; ScrollCode: TScrollCode; var
  ScrollPos: Integer);
begin
  Invalidate;
end;

procedure TrkView.Clear;
begin
  if csDesigning in ComponentState then Exit;
  Items.Clear;
  Selection.Clear;
  sbVert.PageSize := -1;
  sbVert.Position := 0;
  CalcView(True);
  FNrSelected := 0;
  IdxItem:= -1;
  FSelected := -1;
  HotIdx:= -1;
  Invalidate;
  FModified := False;
end;

function TrkView.Scroll(delta, value: integer): integer;
var
  newPos: integer;
begin
  if delta < 0 then begin
    newPos := sbVert.Position + Value;
    if newPos > Range then newPos := Range;
  end else newPos := sbVert.Position - Value;
  sbVert.Position := newPos;
  Result := 1;
end;

function TrkView.GetItemRect(grp, idx: Integer): TRect;
var
  x, y: Integer;
begin
  if Items.Count = 0 then Exit;
  y:= FCellOffset;
  if FCellAuto then
    x:= FCellSpaceX + (Idx mod ViewColumns) * CellColSize
  else
    x:= FCellXOffset + (Idx mod ViewColumns) * ImgWidth;
  y:= y + (Trunc(idx / ViewColumns) * ImgHeight);
  Result:= Rect(x, y, x + FCellWidth, y + FCellHeight -1);
end;

function TrkView.GetItemAtXY(pt: TPoint): Integer;
begin
  Result:= ItemAtXY(pt,(not FCellSelect));
end;

function TrkView.IdxAtXY(pt: TPoint; HitTest: Boolean): integer;
begin
  Result:= ItemAtXY(pt, HitTest);
  if Result > -1 then Result:= Items[Result];
end;

function TrkView.ItemAtXY(pt: TPoint; HitTest: Boolean): integer;
var
  x, y, Idx, xpos, ypos, i, j, k, vOff, xVal: Integer;
  xOk, yOK, Hit : Boolean;
begin
  Result:= -1;
  if Items.Count = 0 then Exit;
  x:= pt.X;
  y:= pt.Y;
  if HeaderVisible then y:= y - FHeaderHeight;
  Idx:= -1;
  FHintColumn:= -1;
  if vsbVisible then vOff:= sbVert.Position else vOff:= 0;
  if FCellAuto then begin
    j:= (CellColSize * ViewColumns) - FCellSpaceX;
    xVal:= FCellSpaceX;
  end else begin
    j:= FCellXOffset + (ViewColumns * ImgWidth);
    xVal:= FCellXOffset;
  end;
  if (x >= xVal) and (x < j) then begin
    if FCellAuto then begin
      xpos:= x div CellColSize; //  in what column is x?
      k := FCellSpaceX + (xpos * CellColSize);
      i := k + FCellWidth;
    end else begin
      xpos := ((x - FCellXOffset) div ImgWidth);
      k := FCellXOffset + (xpos * ImgWidth);
      i := k + ImgWidth - FCellSpaceX;
    end;
    xOK:= (x >= k) and (x <= i);
    i := ViewRows * ImgHeight;
    ypos := (y + (i - (i - (vOff - FCellOffset)))) div ImgHeight;
    i := FCellOffset + (ypos * ImgHeight) + ImgHeight - FCellSpace;
    j := vOff + y;
    yOk:= (j >= FCellOffset) and (j <= i);
    if (xOk and yOk) then begin   // where in cell is mouse ;)
      SelX:= x - k;
      SelY:= i - j;
      Idx := xpos + (ypos * ViewColumns);
      Hit:= true;
      if (HitTest)then
        if (Assigned(FOnCellHit) and (Idx < Items.Count)) then
          FOnCellHit(Self, FMainBmp.Canvas, Items[Idx], SelX, SelY, Hit);
      if (not Hit) then Idx:= -1;
      FHintcolumn:= -1;      // do calc current column
      if FUseAsList then begin
        if hsbVisible then x:= x + sbHorz.Position;
        j:= FCellOffset;
        if (x > FCellOffset) and (x < (FCellWidth) + FCellOffset) then
          for i:= Low(FColumns) to High(FColumns) do begin
            j:= j + FColumns[i];
            if (x < j) and (FHintcolumn = -1) then FHintColumn:= i;
          end;
      end;
    end;
  end;
  if Idx >= Items.Count then Idx:= -1;
  Result:= Idx;
end;

function TrkView.ItemCloseToXY(X, Y: Integer): integer;
var
  xpos, ypos, Selected, i, j, k: integer;
  ColumnOk, RowOk, AtEnd: Boolean;
begin
  xpos := 0;
  ypos := 0;
  ColumnOk := false;
  RowOk := false;
  FMouseAtEnd := false;
  AtEnd := false;

  j := FCellXOffset + (ViewColumns * ImgWidth);

  if (x < FCellXOffset) then x := FCellXOffset;
  if (x >= j) then x := j;

  if (x >= FCellXOffset) and (x <= j) then begin
    xpos := ((x - FCellXOffSet) div ImgWidth);
    i := ((xpos * ImgWidth) + ImgWidth) + FCellXOffset;
    k := FCellXOffset + (xpos * ImgWidth);
    if (x >= k) and (x <= i) then ColumnOk := true else ColumnOk := false;
    if ColumnOk and (x >= (k + (FCellWidth div 2))) and (x <= i) then AtEnd := true;
    if (xpos >= ViewColumns) then begin
      xpos := ViewColumns - 1;
      AtEnd := True;
    end;

    i := ViewRows * ImgHeight;
    ypos := (y + (i - (i - (sbVert.Position - FCellOffset)))) div
      ImgHeight;
    i := FCellOffset + (ypos * ImgHeight) + ImgHeight - FCellSpace;
    j := sbVert.Position + y;
    if (j >= FCellOffset) and (j <= i) then
      RowOk := true
    else
      RowOk := false;
  end;

  Selected := xpos + (ypos * ViewColumns);

  FMouseItem := -1;
  if ((ColumnOk) and (RowOk)) then
  begin
    if (Selected >= Items.Count) then
    begin
      Selected := Items.Count - 1;
      FMouseAtEnd := True
    end
    else
      FMouseAtEnd := AtEnd;
    FMouseItem := Selected;
    Result := Selected
  end
  else
    Result := -1;
end;

procedure TrkView.PaintSimple;
var
  x, y, w, h, cx, cy, nx, ny, fx, fy, d1, scr: integer;
  XPos, YPos, Index, Col, Row, idxLimit, ImageHeight: Integer;
  YCount, bx, by, bdx, bdy, addx, addy, imw, imh: Integer;
  R: TRect;
  slSize, slMain: integer;
  slPnt: PRGB24;
  bool: boolean;
  SItem: TsvItemState;
  i: Integer;
begin
  ScrollOffset:= -FCellOffset;
  if sbVert.Position > Range then sbVert.Position := Range;
  w := ClientWidth - FCtrlSize;
  scr:= w;
  h := ClientHeight;
  if UseAsList then begin
    if (FCellWidth + (2 * FCellOffset)) > w then
      w:= FCellWidth + (2 * FCellOffset) + 1;
    if hsbVisible then h := ClientHeight - pnlHorz.Height;
  end;
  imw := w - 2;
  imh := h - 1;
  FMainBmp.Canvas.Brush.Color := Color;
  FMainBmp.Width := w;
  FMainBmp.Height := h;
  FMainBmp.Canvas.FillRect(Rect(0, 0, imw + 2, imh + 1));
  slMain := Integer(FMainBmp.ScanLine[0]);
  slSize := Integer(FMainBmp.ScanLine[1]) - slMain;
  if vsbVisible then begin
    //YCount := Trunc((sbVert.Position - FCellOffset)/ ImgHeight);
    YCount := ((sbVert.Position - FCellOffset) + FCellSpace) div ImgHeight; // RMK
    YPos := FCellOffset - sbVert.Position + YCount * ImgHeight;
    ScrollOffset := sbVert.Position - FCellOffset;
  end else begin
    YCount := 0;
    YPos:= FCellOffset;
  end;
  Index:= YCount * ViewColumns;
  if hsbVisible then ScrollOffsetX:= sbHorz.Position else ScrollOffsetX:= 0;
  ViewIdx := Index; // First index in view...
  bool := false;
  if ViewColumns <= 0 then ViewColumns:= 1;
  if HeaderVisible then YPos:= YPos + FHeaderHeight;
  if (Index < Items.Count) and (Items.Count > 0) then
    repeat
      if FCellAuto then
        XPos:= FCellSpaceX + (Index mod ViewColumns) * CellColSize
      else
        XPos := FCellXOffset + (Index mod ViewColumns) * ImgWidth;
      R := Rect(XPos, YPos, XPos + FCellWidth, YPos + FCellHeight);
      if R.Bottom > 0 then begin
        if R.Top > Height then Break;
        SItem:= svNormal;
        if (Selection.IndexOf(Items[Index]) <> -1) then SItem:= svSelected;
        //if (Selection.IndexOf(Items[Index]) <> -1) then begin //rmk!
        //  if Index = HotIdx then SItem:= svHotSelected
        //  else SItem:= svSelected;
        //end else
        if Index = HotIdx then SItem:= svHot;
        if IsEditing and (Index = FSelected) then SItem:= svNormal;
        if (FUseAsList) and (Assigned(TMethod(FOnListPaint).Code)) then
          FOnListPaint(Self, FMainBmp.Canvas, R, Index, Items[Index], SItem, FColumns)
        else if Assigned(TMethod(FOnCellPaint).Code) then
          FOnCellPaint(Self, FMainBmp.Canvas, R, Index, Items[Index], SItem);
      end;
      Inc(Index);
      if (Index mod ViewColumns) = 0 then Inc(YPos, ImgHeight);
      bool := (YPos >= Height) or (Index = Items.Count)
    until bool;
  if (UseAsList) and (PaintGrid) and (High(FColumns) > 0)then begin
    FMainBmp.Canvas.Pen.Color:= FLinesColor;
    bx:= FCellOffset;
    for i:= 0 to High(FColumns) do begin
      bx:= bx + FColumns[i];
      FMainBmp.Canvas.MoveTo(bx, 0);
      FMainBmp.Canvas.LineTo(bx, FMainBmp.Height);
    end;
  end;
  // Selection starts here
  if (Selecting) and (SelStart.X <> SelEnd.X) and (SelStart.Y <> SelEnd.Y) then
  begin
    bx := min(SelStart.X, SelEnd.X);
    bdx := max(SelStart.X, SelEnd.X);
    by := min(SelStart.Y, SelEnd.Y);
    bdy := max(SelStart.Y, SelEnd.Y);
    by := by - ScrollOffset;
    bdy := bdy - ScrollOffset;
    if hsbVisible then begin
      if bx < sbHorz.Position then bx := sbHorz.Position;
      if bdx > sbHorz.Position + scr - 1 then bdx := sbHorz.Position + scr - 1;
    end else begin
      if bx < 0 then bx := 0;
      if bdx > imw then bdx := imw;
    end;
    if HeaderVisible then begin
      by:= by + FHeaderHeight;
      bdy:= bdy + FHeaderHeight;
      if by < FHeaderHeight then by:= FHeaderHeight;
    end else if by < 0 then by := 0;
    if bdy > imh then bdy := imh;
    // Smart Alphablend
    row := (slSize * by + slMain) + (bx * 3);
    for y := by to bdy do begin
      slPnt := pRGB24(row);
      for x := bx to bdx do begin
        slPnt.r := Ra[slPnt.r];
        slPnt.g := Ga[slPnt.g];
        slPnt.b := Ba[slPnt.b];
        inc(slPnt)
      end;
      inc(row, slSize);
    end;
    FMainBmp.Canvas.Pen.Width := 1;
    FMainBmp.Canvas.Pen.Color := ColorSel;
    FMainBmp.Canvas.Brush.Style := bsClear;
    FMainBmp.Canvas.Rectangle(Rect(bx, by, bdx + 1, bdy + 1));
  end;
  // Paint Header if needed
  if FHeaderVisible then begin
    R:= Rect(0,0, FMainBmp.Width - 1, FHeaderHeight);
    if (Assigned(TMethod(FOnHeaderPaint).Code)) then
      FOnHeaderPaint(Self, FMainBmp.Canvas, R, FCellXOffset, FHeaderHot, FHeaderState, FColumns);
    if FDragging then begin
      FMainBmp.Canvas.Pen.Color:= FLinesColor;
      x:= FCellOffset;
      for i := 0 to FDragColumn do x:= x + FColumns[i];
      FMainBmp.Canvas.MoveTo(x, 0);
      FMainBmp.Canvas.LineTo(x, FMainBmp.Height);
    end;
  end;

  {
  FMainBmp.Canvas.Pen.Color:= clRed;
  for i:= 0 to ViewColumns do begin
    x:= i * CellColSize;
    FMainBmp.Canvas.MoveTo(x, 0);
    FMainBmp.Canvas.LineTo(x, FMainBmp.Height);
  end;
  }

  if UseAsList then
    BitBlt(Canvas.Handle, 0, 0, scr, FMainBmp.Height,
      FMainBmp.Canvas.Handle, sbHorz.Position, 0, SRCCOPY)
  else
    BitBlt(Canvas.Handle, 0, 0, scr, FMainBmp.Height,
      FMainBmp.Canvas.Handle, 0, 0, SRCCOPY);
end;

function TrkView.HzScrollbar: Boolean;
begin
  Result:= ((FCellWidth + (2 * FCellOffset)) > (ClientWidth - FCtrlSize)) and (UseAsList);
end;

procedure TrkView.CalcViewSimple;
var
  i, n, ViewHeight: Integer;
  hzSB: Boolean;
  Rows: Integer;
  Page, VMax, VPos : Integer;
begin
  if (Items.Count < 1) then Exit;
  FCtrlSize := 0;
  hzSB:= HzScrollbar;
  if FCellAuto then begin
    // New solution ..
    i:= FCellWidth + FCellSpace; //rmk (FCellSpace shl 1);
    n:= ClientWidth div i;
    ViewColumns := n;
    FCellSpaceX:= 1 + ((ClientWidth div ViewColumns) - FCellWidth) shr 1;
    // --- end new solution
  end else begin
    FCellSpaceX:= FCellSpace;
    ViewColumns := Trunc((ClientWidth - ((2 * FCellOffset) - FCellSpaceX)) / ImgWidth);
  end;
  if (ViewColumns = 0) or (UseAsList) then ViewColumns := 1;
  if (HeaderVisible) and (hzSB) then     // Do we have a header and a scrollbar?
    ViewHeight:= ClientHeight - (pnlHorz.Height + FHeaderHeight + (2 * FCellOffset))
  else
  if (HeaderVisible) then                // Only header?
    ViewHeight:= ClientHeight - (FHeaderHeight + (2 * FCellOffset))
  else                               // Neither? Not listview I guess ;)
    ViewHeight:= ClientHeight - (2 * FCellOffset);
  Rows := Round(((ViewHeight / ImgHeight)) + 0.50);
  ViewRows := Rows;
  if (Rows * ImgHeight) > ViewHeight  then dec(Rows);
  vsbVisible:= (ViewColumns * Rows) < Items.Count;
  if vsbVisible then FCtrlSize := pnlVert.Width else FCtrlSize:= 0;
  hsbVisible:= hzSB or HzScrollbar;
  pnlVert.Visible := vsbVisible;
  pnlHorz.Visible := hsbVisible;
  pnlDummy.Visible:= hsbVisible and vsbVisible;
  if vsbVisible then begin
    if FCellAuto then begin
      // New solution ..
      i:= FCellWidth + FCellSpace;//rmk (FCellSpace shl 1);
      n:= (ClientWidth - FCtrlSize) div i;
      ViewColumns := n;
      FCellSpaceX:= 1 + (((ClientWidth - FCtrlSize) div ViewColumns) - FCellWidth) shr 1;
      // --- end new solution
    end else begin
      FCellSpaceX:= FCellSpace;
      ViewColumns := Trunc((ClientWidth - ((2 * FCellOffset) - FCellSpaceX) - FCtrlSize) / ImgWidth);
    end;
  end;
  if (ViewColumns = 0) or (UseAsList) then ViewColumns := 1;
  FCellXOffset:= (ClientWidth - FCtrlSize - ((ViewColumns * ImgWidth) - FCellSpaceX)) shr 1;
  if (hsbVisible) or (not CenterView) then FCellXOffset:= FCellOffset;
  CalcHorzSB; // Update Horizontal scrollbar

  CellColSize:= (ClientWidth - FCtrlSize) div ViewColumns;

  if (not vsbVisible) then Exit;
  if (FHeaderVisible) and (hsbVisible) then     // Do we have a header and a scrollbar?
    ViewHeight:= ClientHeight - (pnlHorz.Height + FHeaderHeight + FCellOffset)
  else
  if (FHeaderVisible) then                // Only header?
    ViewHeight:= ClientHeight - (FHeaderHeight + FCellOffset - FCellSpace)
  else                               // Neither? Not listview I guess ;)
    ViewHeight:= ClientHeight - (FCellOffset - FCellSpace);

  VMax := FCellOffset + (ImgHeight * (Round((Items.Count /  ViewColumns + 0.49))));
  VPos := sbVert.Position;
  Page:= sbVert.PageSize;
  n:= VMax;
  if n < 0 then n:= 0;
  if n > ViewHeight then VMax := n else VMax := 0;
  Range := 1 + VMax - ViewHeight;
  if Range < 0 then begin
    Range := 0;
    VMax := 0;
  end;
  if Page >= VMax then Page := -1 else Page:= ViewHeight + 1;
  if (VPos > Range) and (Range >= 0) then VPos := Range;
  if ImgHeight < 1 then sbVert.SmallChange := 1
  else sbVert.SmallChange := ImgHeight;
  if Page < 1 then sbVert.LargeChange := 1
  else sbVert.LargeChange := Page;
  sbVert.PageSize:= Page;
  // If needed apply to hsb as well
  vsbVisible:= (VMax > 0) and (VMax > VPos);
  pnlVert.Visible := vsbVisible;
  if vsbVisible then sbVert.SetParams(VPos, 0, VMax);
  ImgWidth := FCellWidth + FCellSpaceX;
  ImgHeight := FCellHeight + FCellSpace;
  CellColSize:= (ClientWidth - FCtrlSize) div ViewColumns;
end;

procedure TrkView.CalcHorzSB;
var
  w, delta, nhv, range, page: Integer;
  MinV, MaxV, PosV, WidthSB, LargeV: Integer;
begin
  MaxV:= -1;
  PosV:= -1;
  MinV:= -1;
  if hsbVisible then begin
    w:= ClientWidth - FCtrlSize;
    delta:= (FCellWidth + (2 * FCellOffset)) - w;
    try
      WidthSB:= sbHorz.ClientWidth;
      Page:= sbHorz.PageSize;
      PosV:= sbHorz.Position;
      MinV:= 0;
      MaxV:= w + delta;
      nhv:= MaxV;
      if nhv < 0 then nhv:= 0;
      if nhv > WidthSB then MaxV:= nhv else MaxV:= 0;
      range:= 1 + MaxV - WidthSB;
      if range < 0 then begin
        range:= 0;
        MaxV:= 0;
      end;
      if Page >= MaxV then Page:= -1 else Page:= WidthSB + 1;
      if (PosV > range) and (range >= 0) then PosV:= range;
      if Page < 1 then LargeV:= 1 else LargeV:= MaxV - Page;
      sbHorz.SmallChange := 1;
      sbHorz.LargeChange:= LargeV;
      sbHorz.PageSize:= Page;
      sbHorz.SetParams(PosV, MinV, MaxV);
      maxXScroll:= Delta;
    except
      ShowMessage('Bummer! ' + IntToStr(PosV) + ' ' + IntToStr(MinV) + ' ' + IntToStr(MaxV));
    end;
  end;
end;

procedure TrkView.CalcView(Clear: Boolean);
begin
  ImgWidth := FCellWidth + FCellSpaceX;
  ImgHeight := FCellHeight + FCellSpace;
  if Clear then begin
    sbVert.PageSize:= -1;
    sbHorz.PageSize := -1;
  end;
  CalcViewSimple;
  CalcHorzSB;
end;

procedure TrkView.CalcSelectionSimple;
var
  i: integer;
  Sel, bool: Boolean;
  Cell, R: TRect;
  PixPosY: Integer;
begin
  if Items.Count = 0 then Exit;
  PixPosY:= FCellOffset;
  Selection.Clear;
  bool:= False;
  i:= 0;
  repeat
    // rmk Cell.Left:= FCellXOffset + (i mod ViewColumns) * ImgWidth;
    Cell.Left:= FCellSpaceX + (i mod ViewColumns) * CellColSize;
    Cell.Top:= PixPosY - FCellOffset;
    Cell.Right:= Cell.Left + FCellWidth;
    Cell.Bottom:= Cell.Top + FCellHeight;
    Sel:= IntersectRect(r, SelRect, Cell);
    if (Sel) and (Not CellSelect) then
      if Assigned(TMethod(FOnCellSelect).Code) then
        FOnCellSelect(Self, FMainBmp.Canvas, Cell, Items[i], Sel);
    if Sel then begin
      if not bool then begin
        bool:= True;
        IdxItem:= i;
      end;
      Selection.Add(Items[i]);
    end;
    Inc(i);
    if (i < Items.Count) and (i mod ViewColumns = 0) then
      Inc(PixPosY, (FCellHeight + FCellSpace));
  until (i = Items.Count) or (PixPosY > SelRect.Bottom + FCellHeight);
end;

procedure TrkView.CalcSelection;
begin
  SelRect.Left:= Min(SelStart.X, SelEnd.X);
  SelRect.Right:= Max(SelStart.X, SelEnd.X);
  SelRect.Top:= Min(SelStart.Y, SelEnd.Y);
  SelRect.Bottom:= Max(SelStart.Y, SelEnd.Y);
  CalcSelectionSimple;
  if Assigned(TMethod(FOnSelecting).Code) then FOnSelecting(Self, Selection.Count);
end;

procedure TrkView.WMERASEBKGND(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TrkView.WMSETFOCUS;
begin
  inherited;
  if FCanFocused then begin
    FFocused := True;
    Invalidate;
  end;
end;

procedure TrkView.WMKILLFOCUS;
begin
  inherited;
  if FCanFocused then begin
    FFocused := False;
    Invalidate;
  end;
end;

procedure TrkView.SetCanFocused;
begin
  FCanFocused := Value;
  if FCanFocused then TabStop := True else TabStop := False;
end;

procedure TrkView.SetSelColor;
var
  C: LongInt;
  i: Integer;
begin
  C:= ColorToRGB(Value);
  FColorSelection.B:= Byte(C shr 16);
  FColorSelection.G:= Byte(C shr 8);
  FColorSelection.R:= Byte(C);
  for i:= 0 to 255 do begin
    Ra[i] := (FColorSelection.R - i) shr 3 + i;
    Ga[i] := (FColorSelection.G - i) shr 3 + i;
    Ba[i] := (FColorSelection.B - i) shr 3 + i;
  end;
end;

function TrkView.GetSelColor;
begin
  Result := (FColorSelection.B shl 16) + (FColorSelection.G shl 8) +
    FColorSelection.R;
end;

function TrkView.IsFocused;
begin
  Result := (FFocused) and (FCanFocused);
end;

procedure TrkView.SimpleMouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  sel, i, j, k: Integer;
begin
  if (Button = mbLeft) or (Button = mbRight) then begin
    Sel:= ItemAtXY(Point(X, Y), (not FCellSelect));
    if Sel <> -1 then begin
      if FMulti then begin
        if (ssShift in shift) and not (ssCtrl in shift) then begin
          Selection.Clear;
          i := Sel;
          j := FSelected;
          if j = -1 then j := 0;
          if j < i then begin
            k := j;
            j := i;
            i := k;
          end;
          for k := i to j do Selection.Add(Items[k]);
        end;
        if (ssCtrl in shift) and not (ssShift in shift) then begin
          i:= Selection.IndexOf(Items[sel]);
          if i <> -1 then Selection.Delete(i) else Selection.Add(Items[sel]);
        end;
        if (not (ssCtrl in shift)) and (not (ssShift in shift)) then
        begin
          if Selection.IndexOf(Items[Sel]) = -1 then begin
            Selection.Clear;
            Selection.Add(Items[Sel]);
          end;
        end;
      end else begin
        Selection.Clear;
        Selection.Add(Items[Sel]);
      end;
      FSelected:= Sel;
      SetInView(-1, FSelected);
      Invalidate;
    end else begin
      Selection.Clear;
      FSelected := -1;
      FPrevSel.Top := -1;
      FPrevSel.Left := -1;
      FPrevSel.Right := -1;
      FPrevSel.Bottom := -1;
      SelStart.X := X;
      SelStart.Y := Y + ScrollOffset;
      if FHeaderVisible then SelStart.Y:= SelStart.Y - FHeaderHeight;
      SelEnd := SelStart;
      Selecting := FMulti;
      viewDeltaScroll := 0;
      viewDeltaScrollX := 0;
      SelTimer.Enabled := True;
      Invalidate;
    end;
    if Assigned(FOnSelecting) then FOnSelecting(Self, Selection.Count);
  end;
  FLastIndex := FSelected;
end;

procedure TrkView.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if (CanFocus) then SetFocus;
  if Items.Count = 0 then Exit;
  FDownX:= x;
  FDownY:= y;
  FDownTick:= GetTickCount;

  SelTimer.Enabled:= False; //rmkidag
  SCPos:= sbVert.Position;
  if ScrollMode then Exit;

  if (FHeaderVisible) and (y < FHeaderHeight) and (y >= 0) then
  begin
    InHeader:= True;
    if (FDragColumn <> -1) then begin
      FDragging:= True;
      FDragStartX:= x;
      FDragStartSize:= FColumns[FDragColumn];
    end;
    if (not FDragging) and (FHeaderHot <> -1) and (Button = mbLeft) then
    begin
      FHeaderState:= svSelected;
      if Assigned(TMethod(FOnHeaderClick).Code) then
        OnHeaderClick(Self, FHeaderHot);
    end;
    Invalidate;
    Exit;
  end else
    InHeader:= False;
  if dblClicked then Exit;
  if (UseAsList) and (hsbVisible) then x:= x + sbHorz.Position;
  SimpleMouseDown(Button, Shift, X, Y);
  y:= y + YAdjOff;
  if Assigned(OnMouseDown) then OnMouseDown(Self, Button, Shift, X, Y);
end;

procedure TrkView.UpdateView;
begin
  if Items.Count > 0 then SetInView(-1, ViewIdx);
end;

procedure TrkView.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y:
  Integer);
var
  dwTicks, dwElapsed: DWord;
begin
  FDragging:= False;
  FDragColumn:= -1;
  SelTimer.Enabled := false;
  Selecting := false;
  SelScroll := false;
  if dblClicked then dblClicked := false;

  if ScrollMode then begin
    dwTicks := GetTickCount;
    if dwTicks >= FDownTick then
      dwElapsed := dwTicks - FDownTick
    else
      dwElapsed := $FFFFFFFF - FDownTick + dwTicks;
    SCUp:= SCY < 0;
    SCValue:= Abs(SCY);
    SelTimer.Enabled:= (dwElapsed < 175);
    Exit;
  end;

  Invalidate;
  if Assigned(OnMouseUp) then OnMouseUp(Self, Button, Shift, X, Y);
end;

procedure TrkView.CMHintShow(var Message: TCMHintShow);
var
  InfoTip :string;
  Item, i, j: Integer;
  R: TRect;
  bool: Boolean;
begin
  if Assigned(FOnHintShow) then begin
    Item:= ItemAtXY(Message.HintInfo.CursorPos, False);
    if (Item <> -1) then begin
      R:= GetItemRect(-1, Item);
      Item:= Items[Item];
    end else Exit;
    if vsbVisible then R.Top:= (R.Top - sbVert.Position) + 1
    else R.Top:= R.Top + 1;
    R.Bottom:= (R.Top + FCellHeight);
    if FHeaderVisible then begin
      R.Top:= R.Top + FHeaderHeight;
      R.Bottom:= R.Top + FCellHeight;
      i:= 0;
      j:= FcellOffset;
      while i < FHintColumn do begin
        j:= j + FColumns[i];
        i:= i + 1;
      end;
      R.Left:= j;
      R.Right:= j + FColumns[i];
    end;
    bool:= False;
    OnHintShow(Self, Item, FHintColumn, InfoTip, bool);
    bugStr:= 'Bummer: ' + IntToStr(FHintColumn);
    if (not bool) or (InfoTip = '') or (Selecting) or (FDragging) or (IsEditing) then
      Message.Result := -1
    else begin
      Message.HintInfo.HintStr := InfoTip;
      Message.HintInfo.CursorRect:= R;
      Message.HintInfo.HideTimeout := 120000;
      Message.Result := 0;
    end;
  end
  else inherited;
end;

procedure TrkView.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Idx: Integer;
  ViewHeight, i, n, j, k: Integer;
  bool, dummy: Boolean;
  s: String;
begin
  if dblClicked then dblClicked := false;

  if ScrollMode then begin
    if (ssLeft in Shift) then begin
      SCY:= FDownY - Y;
      i:= SCPos + SCY;
      if i < sbVert.Min then sbVert.Position:= sbVert.Min else
      if i > sbVert.Max then sbVert.Position:= sbVert.Max else
      sbVert.Position:= i;
      Invalidate;
    end;
    Exit;
  end;

  bool:= False;
  if (FHeaderState = svSelected) and (ssLeft in Shift) then Exit;
  if not (ssLeft in Shift) then Selecting := false;
  if Selecting then begin
    if FHeaderVisible then begin
      y:= y - FHeaderHeight;
      if x < 0 then
        viewDeltaScrollX := x * abs(x)
      else if x > (ClientWidth - FCtrlSize) then
        viewDeltaScrollX := (x - (ClientWidth - FCtrlSize)) * (x - (ClientWidth - FCtrlSize))
      else
        viewDeltaScrollX := 0;
      if (viewDeltaScrollX <> 0) and (ssShift in Shift) then
        viewDeltaScrollX := viewDeltaScrollX * 2;
    end;
    if (FHeaderVisible) and (hsbVisible) then
      ViewHeight:= ClientHeight - (pnlHorz.Height + FHeaderHeight)
    else
    if (FHeaderVisible) then ViewHeight:= ClientHeight - FHeaderHeight
      else ViewHeight:= ClientHeight;
    if y < 0 then viewDeltaScroll := y * abs(y)
    else
    if y > ViewHeight then
      viewDeltaScroll := (y - ViewHeight) * (y - ViewHeight)
    else viewDeltaScroll := 0;
    if (viewDeltaScroll <> 0) and (ssShift in Shift) then
      viewDeltaScroll := viewDeltaScroll * 2;
    SelEnd.X := X + ScrollOffsetX;
    SelEnd.Y := Y + ScrollOffset;
    CalcSelection;
    bool:= True;
  end else
  if (FHeaderVisible) and (FDragging) then begin
    i:= FDragStartSize + (x - FDragStartX);
    if i > 2 then FColumns[FDragColumn]:= i else FColumns[FDragColumn]:= 2;
    for i := 0 to High(FColumns) do s:= s + IntToStr(FColumns[i]) + ',';
    j:= Length(s);
    SetLength(s, j-1);
    Columns:= s;
    FCellWidth:= FListWidth;
    CalcViewSimple;
    bool:= True;
  end else begin
    if HotTracking then begin
      if hsbVisible then x:= x + sbHorz.Position;
      if (FHeaderVisible) and (y < FHeaderHeight) and (y >= 0) then begin
        if High(FColumns) > 0 then begin
          j:= 0;
          k:= 0;
          x:= x - FCellXOffset;
          dummy:= False;
          FDragColumn:= -1;
          n:= FHeaderHot;
          FHeaderHot:= -1;
          FHeaderState:= svNormal;
          while j <= High(FColumns) do begin
            k:= k + FColumns[j];
            if (abs(x - k) < 5) and (not Dummy) then begin
              FDragColumn:= j;
              FDragOffs:= (x - k);
              Self.Cursor:= crHSplit;
              dummy:= True;
            end;
            if (x > 0) and (x < FCellWidth) then begin
              if (x <= k) then begin
                FHeaderHot:= j;
                FHeaderState:= svHot;
                j:= High(FColumns);
              end;
            end;
            j:= j + 1;
          end;
          bool:= (n <> FHeaderHot);
          if (not dummy) then begin
            Self.Cursor:= crDefault;
            FDragColumn:= -1;
          end;
        end;
      end else begin
        Self.Cursor:= crDefault;
        if (FHeaderHot <> -1) then begin
          FHeaderHot:= -1;
          FHeaderState:= svNormal;
          bool:= True;
        end;
        Idx:= ItemAtXY(Point(x, y), (not FCellSelect));
        if Idx <> HotIdx then begin
          HotIdx:= Idx;
          bool:= True;
        end;
      end;
    end;
    if Assigned(OnMouseMove) then OnMouseMove(Self, Shift, x, y);
  end;
  if bool then Invalidate;
end;

procedure TrkView.CMMouseLeave(var Message: TMessage);
begin
  FHeaderHot:= -1;
  FHeaderState:= svNormal;
  HotIdx:= -1;
  Invalidate;
end;

procedure TrkView.DblClick;
begin
  FDownTick:= GetTickCount;
  if IsEditing then Exit;
  dblClicked := True;
  if (Selection.Count > 0) and (not InHeader) then
    if Assigned(OnDblClick) then
      OnDblClick(Self);
end;

procedure TrkView.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWord = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TrkView.OnSelScroll(Sender: TObject);
var
  i: integer;
  bool: Boolean;
begin
  if ScrollMode then begin
    if SCUp then i:= -SCValue else i:= SCValue;
    i:= sbVert.Position + i;
    if i < sbVert.Min then sbVert.Position:= sbVert.Min else
    if i > sbVert.Max then sbVert.Position:= sbVert.Max else
    sbVert.Position:= i;
    Invalidate;
    if SCValue <> 0 then SCValue:= Trunc((SCValue - Ln(SCValue)) / 1.1);
    SelTimer.Enabled:= SCValue > 0;
    Exit;
  end;

  bool:= false;
  if (viewDeltaScroll <> 0) then
  begin
    if viewDeltaScroll < 0 then
      SelEnd.Y := ScrollOffset
    else
      SelEnd.Y := ScrollOffset + ClientHeight;
    SelEnd.Y := SelEnd.Y + viewDeltaScroll;

    if vsbVisible then begin
      i := sbVert.Position + viewDeltaScroll;
      if i <> sbVert.Position then begin
        sbVert.Position := i;
        bool:= true;
      end;
    end;
  end;

  if (hsbVisible) and (viewDeltaScrollX <> 0) then begin
    if viewDeltaScrollX < 0 then
      SelEnd.X:= ScrollOffsetX
    else
      SelEnd.X:= ScrollOffsetX + ClientWidth;

    i := sbHorz.Position + viewDeltaScrollX;
    if i <> sbHorz.Position then begin
      if i >= maxXScroll then i:= maxXScroll;
      sbHorz.Position:= i;
      bool:= true;
    end;
  end;

  if bool then PostMessage(Handle, CM_REFRESH, 0, 0);

  CalcSelection;
  SelTimer.Enabled := Selecting;
end;

procedure TrkView.AlignControls(AControl: TControl; var Rect: TRect);
begin
  inherited AlignControls(AControl, Rect);
end;

function TrkView.SimpleDoKey(var Key: Word; Shift: TShiftState): Boolean;
var
  idx, old, i, j, k, maxInt: Integer;
  DoInvalid, Done: Boolean;
begin
  j := sbVert.Position;
  maxInt := Range;
  idx:= -1;
  old:= -1;
  Done:= True;
  DoInvalid := False;
  if (FSelected = -1) and (Selection.Count = 0) then begin
    if (Shift = []) then begin
      FSelected := 0;
      FLastIndex := 0;
      DoInvalid := true;
    end;
  end else
  begin
    idx:= FSelected;
    old:= idx;
    case Key of
      vk_Home:
        begin
          if (ssCtrl in Shift) then
          begin
            sbVert.Position := 0;
            DoInvalid := True;
          end;
        end;
      vk_End:
        begin
          sbVert.Position := sbVert.Max;
          DoInvalid := True;
        end;
      vk_Next:
        begin
          i := j + ClientHeight;
          if i > sbVert.Max then
            sbVert.Position := maxInt
          else
            sbVert.Position := i;
          DoInvalid := True;
        end;
      vk_Prior:
        begin
          i := j - ClientHeight;
          if i < 0 then
            sbVert.Position := 0
          else
            sbVert.Position := i;
          DoInvalid := True;
        end;
      vk_Down:  if idx + ViewColumns < Items.Count then idx:= idx + ViewColumns;
      vk_Up:    if idx - ViewColumns >= 0 then idx:= idx - ViewColumns;
      vk_Left:  if idx - 1 >= 0 then idx:= idx - 1;
      vk_Right: if idx + 1 < Items.Count then idx := idx + 1;
      vk_Return:DblClick;
    else
      Done:= False;
    end;
  end;
  if (idx <> old) and (idx <> -1) then begin
    Selection.Clear;
    FSelected:= idx;
    SetInView(-1, FSelected);
    if (ssShift in Shift) then begin
      if FLastIndex <> -1 then begin
        j := Min(FSelected, FLastIndex);
        k := Max(FSelected, FLastIndex);
        for i := j to k do Selection.Add(Items[i]);
      end;
    end else Selection.Add(Items[idx]);
    DoInvalid := True;
  end;
  if (not (ssShift in Shift)) then FLastIndex := FSelected;
  if DoInvalid then Invalidate;
  if Assigned(FOnSelecting) then FOnSelecting(Self, Selection.Count);
  Result:= Done;
end;

function TrkView.TextWidth(text: String): Integer;
begin
  Result:= FMainBmp.Canvas.TextWidth(text);
end;

function TrkView.GetColumnSize(Column: Integer): Integer;
begin
  if (Column >= Low(FColumns)) and (Column <= High(FColumns))then
    Result:= FColumns[Column] else Result:= -1;
end;

procedure TrkView.SetInView(grp, idx: Integer);
var
  y, i, h, t: Integer;
  R: TRect;
begin
  YAdjOff:= 0;
  i:= idx;
  if i < 0 then Exit;
  h:= ClientHeight;
  if hsbVisible then h:= (h - pnlHorz.Height) - 1;
  R:= GetItemRect(-1, idx);
  if FHeaderVisible then begin
    R.Top:= R.Top + (FHeaderHeight - FCellSpace);
    R.Bottom:= (R.Bottom + (FHeaderHeight - FCellSpace));
    t:= FHeaderHeight
  end else t:= 0;
  if R.Top < (sbVert.Position + t) then begin
    YAdjOff:= sbVert.Position;
    sbVert.Position:= (R.Top - FCellSpace) - t;
    YAdjOff:= YAdjOff - sbVert.Position;
  end;
  y:= (sbVert.Position + h);
  if R.Bottom > y then begin
    YAdjOff:= sbVert.Position;
    sbVert.Position:= (R.Bottom - h) + 2;
    YAdjOff:= YAdjOff - sbVert.Position;
  end;
  Invalidate;
end;

procedure TrkView.SetAtTop(grp, idx: Integer);
var
  i, t: Integer;
  R: TRect;
begin
  YAdjOff:= 0;
  CalcView(True);
  i:= idx;
  if i < 0 then Exit;
  R:= GetItemRect(-1, idx);
  if FHeaderVisible then begin
    R.Top:= R.Top + (FHeaderHeight - FCellSpace);
    R.Bottom:= (R.Bottom + (FHeaderHeight - FCellSpace));
  end;
  if FHeaderVisible then t:= FHeaderHeight else t:= 0;
  YAdjOff:= sbVert.Position;
  sbVert.Position:= (R.Top - FCellSpace) - t;
  YAdjOff:= YAdjOff - sbVert.Position;
  Invalidate;
end;

procedure TrkView.KeyDown(var Key: Word; Shift: TShiftState);
var
  Done: Boolean;
begin
  if (Items.Count = 0) then begin
    if Assigned(OnKeyDown) then OnKeyDown(Self, Key, Shift);
    Exit;
  end;
  Done:= SimpleDoKey(Key, Shift);
  if (not Done) and Assigned(OnKeyDown) then OnKeyDown(Self, Key, Shift);
end;

procedure TrkView.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
end;

procedure TrkView.WMKeyDown(var Message: TWMKeyDown);
begin
  with Message do KeyDown(CharCode, KeyDataToShiftState(KeyData));
end;

procedure TrkView.WMKeyUp(var Message: TWMKeyDown);
begin
  with Message do KeyUp(CharCode, KeyDataToShiftState(KeyData));
end;

procedure Register;
begin
  RegisterComponents('SmartView', [TrkView]);
end;

end.
